import warnings
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.metrics import classification_report, confusion_matrix
from sklearn.neighbors import KNeighborsClassifier
from sklearn.preprocessing import StandardScaler, LabelEncoder

from utilities import ordinal_encode, normalize, plot_confusion_matrix, plot_counts, import_data, plot_roc_curve

warnings.filterwarnings("ignore")

# Constants
y_data = 'class'
y_values = ['normal', 'anomaly']
df_train, df_test, model_name = import_data(
    train_file_path="nsl-kdd-dataset/" + "KDDTrain+.arff",
    test_file_path="nsl-kdd-dataset/" + "KDDTest+.arff",
    model_name="KNN")

sc = StandardScaler()
enc = LabelEncoder()

# Normalize data
ordinal_encode(df=df_train, categories=y_values, target=y_data)
ordinal_encode(df=df_test, categories=y_values, target=y_data)

normalize(df_train, df_test, y_data, sc, enc)

# Plot absolute quantities of class 0 and class 1
sns.countplot(x=y_data, data=df_train)
plot_counts(model_name=model_name)

# Separate X and y
X_train = df_train.select_dtypes(include=[np.number]).drop(columns = [y_data])
X_test = df_test.select_dtypes(include=[np.number]).drop(columns = [y_data])
y_train = df_train[[y_data]]
y_test = df_test[[y_data]]

# Training model
model = KNeighborsClassifier()
model.fit(X_train, y_train)
y_prediction = model.predict(X_test)
print("Classification report: \n", classification_report(y_test, y_prediction))
plot_confusion_matrix(confusion_matrix=confusion_matrix(y_test, y_prediction),
                      accuracy=model.score(X_test, y_test),
                      model_name=model_name)

# Calculate prediction probabilities for ROC curve
y_score = model.predict_proba(X_test)[:, 1]
plot_roc_curve(y_test, y_score, model_name=model_name)

def predict(prediction_input):
    if len(prediction_input) == 0:
        return
    input_data = pd.DataFrame(prediction_input, columns=X_train.columns)
    return model.predict(input_data)